/*
 *  Usage:  count [-c counter] [-t type] [-s size(WxH)] [-l length] [-u url]
 *
 * Assumed to be CGI script using GET method, therefore will parse QUERY_STRING
 * to find the following args:
 *
 *   counter	counter value to display
 *   type	type subdirectory where to find digits
 *   size	Width x Height of GIF output
 *   length	Number of digits for output, including padding
 *		(default is no padding)
 *   url        URL to look up (default from HTTP_REFERER)
 *
 * Command line args override CGI GET method QUERY_STRING args.
 *
 * Written by Brian Kolaci (bk@galaxy.net)  2-14-96
 *
 * brk  3-10-96
 *     Updated counter to *not* rely on correct SERVER_NAME and
 *     SERVER_PORT being set.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ndbm.h>
#include "gd.h"

#define TYPE_DIR	"/export/www/digits"
#define DEFAULT_TYPE	"default"

void		parseCGI();
void		parseArgs(int, char**);
void		printGif();
void		findURLCount();
gdImagePtr	readImage( char digit );

/*
 * global options
 */
char *	Digits = NULL;
char *	URL = NULL;
char	DigitType[256];
int	OutputHeight =	-1;		/* use largest size in font */
int	OutputWidth =	-1;		/* use size in font */
int	DigitWidth =	5;		/* specify 0 for no padding */

/*
 * Globals
 */

/* IndexNames specifies the list of index names to use
   when looking up a directory. */
char *	IndexNames[] = {
	"index.html",
	"Welcome.html",
	"welcome.html",
	"index.shtml",
	NULL
};

int main( int ac, char** av )
{
	char* env;

	/* HACK, fix bug in netscape to stop caching this counter */
	utime(av[0], NULL);

	/* Make sure we give up in a reasonable time */
	alarm(120);

	strcpy(DigitType, DEFAULT_TYPE);
	DigitType[sizeof(DigitType) - 1] = '\0';

	/* Parse CGI first */
	parseCGI();

	/* Parse command line */
	parseArgs(ac,av);

	if (Digits == NULL)
		findURLCount();

	printGif();

	if (Digits)
		free(Digits);
	if (URL)
		free(URL);
	return(0);
}

/*
 * parseCGI()
 *
 * parses QUERY_STRING and sets values.
 * THIS IS VERY RUDIMENTARY AND SIMPLE.  TRUE CGI PARSING IS NOT DONE
 * HERE.  NO ESCAPE SEQUENCES OR FANCY CHECKING IS DONE.
 */
void parseCGI()
{
	char* p = getenv("QUERY_STRING");
	while (p && *p) {
		/*
		 * Find end of this component and terminate it
		 */
		char * end = strchr(p, '&');
		if (end)
			*end = '\0';

		/*
		 * Process each param
		 */
		if (!strncasecmp(p, "type=", 5)) {
			strncpy(DigitType, p+5, sizeof(DigitType) - 1);
		} else if (!strncasecmp(p, "size=", 5)) {
			char* ht = strchr(p,'x');
			OutputWidth = atoi(p+5);
			if (ht)
				OutputHeight = atoi(++ht);
		} else if (!strncasecmp(p, "length=", 7)) {
			DigitWidth = atoi(p+7);
		} else if (!strncasecmp(p, "counter=", 8)) {
			if (Digits)
				free(Digits);
			Digits = strdup(p+8);
		} else if (!strncasecmp(p, "url=", 4)) {
			if (URL)
				free(URL);
			URL = strdup(p+4);
		}
		if (end)
			p = ++end;
		else
			break;
	}
}

void parseArgs( int ac, char** av )
{
	extern char* optarg;
	extern int   optind;
	int c;

	/* Allow command line override */
	while ((c = getopt(ac, av, "c:t:s:l:u:")) != EOF) {
		switch (c) {
		case 'c':
			if (Digits)
				free(Digits);
			Digits = strdup(optarg);
			break;
		case 't':
			strncpy(DigitType, optarg, sizeof(DigitType) - 1);
			break;
		case 's':
			OutputWidth = atoi(optarg);
			optarg = strchr(optarg,'x');
			if (optarg)
				OutputHeight = atoi(++optarg);
			break;
		case 'l':
			DigitWidth = atoi(optarg);
			break;
		case 'u':
			if (URL)
				free(URL);
			URL = strdup(optarg);
			break;
		}
	}
}

void	printGif()
{
	int		i;
	int		resize = 0;
	int		width = 0;
	int		height = 0;
	char *		p;
	gdImagePtr	digits[10] = {NULL};
	gdImagePtr	out;
	char		dir[2048];

	/*
	 * Load required digits
	 * Calculate size of output image too
	 */

	/* Go to digits directory */
	strcpy(dir, TYPE_DIR);
	strcat(dir, "/");
	strcat(dir, DigitType);
	if (chdir(dir))
		return;

	/* Load zero, for padding */
	if (DigitWidth < strlen(Digits))
		DigitWidth = strlen(Digits);
	if (DigitWidth > strlen(Digits)) {
		digits[0] = readImage('0');
		height = gdImageSY(digits[0]);
		width = gdImageSX(digits[0]) * (DigitWidth - strlen(Digits));
	}

	/* Load rest of required digits */
	for (p = Digits; p && *p; p++) {
		if (isdigit(*p)) {
			int dig = *p - '0';
			if (digits[dig] == NULL) 
				if ((digits[dig] = readImage(*p)) == NULL)
					continue;
			if (gdImageSY(digits[dig]) > height)
				height = gdImageSY(digits[dig]);
			width += gdImageSX(digits[dig]);
		}
	}

	/* Create output image */
	out = gdImageCreate(width, height);
	width = 0;

	/* Draw leading zeros */
	if (digits[0] != NULL) {
		for (i = 0; i < DigitWidth - strlen(Digits); i++) {
			gdImageCopyResized(out, digits[0], width, 0, 0, 0, 
						digits[0]->sx, height,
						digits[0]->sx, digits[0]->sy);
			width += gdImageSX(digits[0]);
		}
	}

	/* Draw rest of digits */
	for (p = Digits; p && *p; p++) {
		if (isdigit(*p)) {
			int dig = *p - '0';
			if (digits[dig] == NULL)
				continue;
			gdImageCopyResized(out, digits[dig], width, 0, 0, 0, 
						digits[dig]->sx, height,
						digits[dig]->sx, digits[dig]->sy);
			width += gdImageSX(digits[dig]);
		}
	}

	/* Make output interlaced */
	gdImageInterlace(out, 1);

	/* Put Expires header to expire now */
	{
	char buf[256];
	time_t t = time(0L);
	struct tm* tms = gmtime(&t);
	strftime(buf, sizeof(buf), "%A, %d-%b-%y %T", tms);
	fprintf(stdout, "Date: %s GMT\n", buf);
	/*fprintf(stdout, "Expires: %s GMT\n", buf);*/
	fprintf(stdout, "Expires: Monday, 30-Mar-70 02:57:45 GMT\n", buf);
	/*fprintf(stdout, "Last-Modified: %s GMT\n", buf);*/
	}

	/* Spit out image */
	fprintf(stdout, "Pragma: no-cache\n");
	fprintf(stdout, "Content-type: image/gif\n\n");
	gdImageGif(out, stdout);

	/* Cleanup */
	gdImageDestroy(out);
	for (i = 0; i < 10; i++) {
		if (digits[i] != NULL)
			gdImageDestroy(digits[i]);
	}
}

gdImagePtr readImage( char digit )
{
	char file[256];
	gdImagePtr im = NULL;
	FILE* fp;

	sprintf(file, "%c.gif", digit);
	if (fp = fopen(file, "r")) {
		im = gdImageCreateFromGif(fp);
		fclose(fp);
	}
	return im;
}

void findURLCount()
{
	char * env;
	char * file = getenv("URL_COUNT_DB");

	/*
	 * Get URL if we don't have it.  Make sure referer is local
	 * local to this server.
	 */
	if (URL == NULL) {
		char* refer = getenv("HTTP_REFERER");
		if (refer) {
			char* p = refer;
			if (!strncmp(p, "http://", 7))
				p += 7;

			/* Skip to the URL */
			while (*p && *p != '/')
				p++;

			/* We should now be at the URL */
			if (*p == '/')
				URL = strdup(p);
		}
	}

	/*
	 * Now lookup URL in counter database
	 * If we can access the database file and found the count,
	 * then return.  Otherwise we drop down and set the default values.
	 */
	if (URL && file) {
		char dbmfile[2048];
	
		strcpy(dbmfile, file);
		strcat(dbmfile, ".pag");

		/*
		 * Try to find the URL in the database
		 */
		if (!access(file, R_OK)) {
			if (searchRegular(file))
				return;
		} else if (!access(dbmfile, R_OK)) {
			if (searchDbm(file))
				return;
		}

		/*
		 * Not found, try using a full filename
		 */
		if (env = getenv("DOCUMENT_ROOT")) {
			char newurl[2048];
			strcpy(newurl, env);
			strcat(newurl, URL);
			free(URL);
			URL = strdup(newurl);
			if (!access(file, R_OK)) {
				if (searchRegular(file))
					return;
			} else if (!access(dbmfile, R_OK)) {
				if (searchDbm(file))
					return;
			}
		}
	}

	/*
	 * If no URL found, just set count from URL_COUNT
	 */
	/* Default digits to display */
	if (env = getenv("URL_COUNT"))
		Digits = strdup(env);
	else
		Digits = strdup("0");
}

/*
 * File locking
 */

int lockfile( int fd )
{
	int ret;
	struct flock lock = { F_WRLCK, 0, 0, 0 };

	while ((ret = fcntl(fd, F_SETLKW, &lock)) < 0 && errno == EINTR)
		continue;

	return(ret);
}
int unlockfile( int fd )
{
	struct flock unlock = { F_UNLCK, 0, 0, 0 };
	return(fcntl(fd, F_SETLKW, &unlock));
}

int searchRegular( const char* file )
{
	int i;
	int ret = 0;
	int isdir = (URL[strlen(URL) - 1] == '/');
	FILE* fp;

	if (fp = fopen(file, "r")) {
		char buf[2048];
		char urlbuf[2048];
		char * p, * q;
		int urlbuflen;

		lockfile(fileno(fp));

		strcpy(urlbuf, URL);
		/*strcat(urlbuf, "/"); */
		urlbuflen = strlen(urlbuf);

		while (fgets(buf, sizeof(buf), fp)) {
			if (!strncmp(buf, URL, strlen(URL))) {
				if (isdir) {	
					for (i = 0; IndexNames[i] != NULL; i++) {
						urlbuf[urlbuflen] = '\0';
						strcat(urlbuf, IndexNames[i]);
						if (!strncmp(buf, urlbuf, strlen(urlbuf)) &&
						    isspace(buf[strlen(urlbuf)]))
							break;
					}
					if (IndexNames[i] == NULL)
						continue;
				}

				/* Found the entry, get the count */
				p = buf + strlen(URL);
				while (!isspace(*p))
					p++;
				/* Skip whitespace */
				while (isspace(*p))
					p++;
				/* Skip leading zeros */
				while (*p == '0')
					p++;
				/* Now find end of string and make a copy */
				q = p;
				while (isdigit(*q))
					q++;
				*q = '\0';
				Digits = strdup(p);
				ret = 1;
				break;
			}
		}
		unlockfile(fileno(fp));
		fclose(fp);
	}
	return(ret);
}

int searchDbm( char* file )
{
	int ret = 0;
	int isdir = (URL[strlen(URL) - 1] == '/');
	DBM* dbm;
	datum d, q;
	char * p;
	int i;

	if (dbm = dbm_open(file, O_RDONLY, 0644)) {
		lockfile(dbm_dirfno(dbm));

		if (isdir) {
			int urlbuflen;
			char urlbuf[2048];
			strcpy(urlbuf, URL);
			urlbuflen = strlen(urlbuf);

			for (i = 0; IndexNames[i] != NULL; i++) {
				urlbuf[urlbuflen] = '\0';
				strcat(urlbuf, IndexNames[i]);

				q.dptr = urlbuf;
				q.dsize = strlen(urlbuf);
				d = dbm_fetch(dbm, q);
				if (d.dptr != NULL)
					break;
			}
		} else {
			q.dptr = URL;
			q.dsize = strlen(URL);
			d = dbm_fetch(dbm, q);
		}

		/* Get count out */
		if (p = d.dptr) {
			while (!isspace(*p))
				p++;
			*p = '\0';
			Digits = strdup(d.dptr);
			ret = 1;
		}

		/* Cleanup and close */
		unlockfile(dbm_dirfno(dbm));
		dbm_close(dbm);
	}
	return(ret);
}

